@@ -39,6 +39,7 @@ import uuid
 
 import cord.exceptions
 from cord.configs import CordConfig
+from cord.constants.model import *
 from cord.http.querier import Querier
 from cord.http.utils import upload_to_signed_url, upload_to_signed_url_list
 from cord.orm.api_key import ApiKeyMeta
@@ -49,9 +50,9 @@ from cord.orm.label_row import LabelRow
 from cord.orm.labeling_algorithm import (
     LabelingAlgorithm, ObjectInterpolationParams
 )
-from cord.orm.model import Model, ModelInferenceParams
+from cord.orm.model import Model, ModelRow, ModelInferenceParams, ModelTrainingParams, ModelOperations
 from cord.orm.project import Project
-from cord.utils.str_constants import *
+from cord.constants.string_constants import *
 
 # Logging configuration
 logging.basicConfig(stream=sys.stdout,
@@ -292,7 +293,7 @@ class CordClientProject(CordClient):
         Save existing label row.
 
         If you have a series of frame labels and have not updated answer
-        dictionaries, call the construct_answer_dictionaries utils function
+        dictionaries, call the construct_answer_dictionaries utilities function
         to do so prior to saving labels.
 
         Args:
@@ -337,6 +338,62 @@ class CordClientProject(CordClient):
         """
         return self._querier.basic_put(LabelRow, uid=uid, payload=None)
 
+    def create_model_row(self,
+                         title=None,
+                         description=None,
+                         features=None,
+                         model=None,
+                         ):
+        """
+        Create a model row.
+
+        Args:
+            title: Model title.
+            description: Model description.
+            features: List of feature feature uid's (hashes) to be included in the model.
+            model: Model (resnet18, resnet34, resnet50,
+                resnet101, resnet152, vgg16, vgg19, yolov5, faster_rcnn, mask_rcnn).
+
+        Returns:
+            ModelRow: A model_hash (uid) string.
+
+        Raises:
+            AuthenticationError: If the project API key is invalid.
+            AuthorisationError: If access to the specified resource is restricted.
+            ModelFeaturesInconsistentError: If a feature type is different than what is supported by the model (e.g. if
+                creating a classification model using a bounding box).
+        """
+        if title is None:
+            raise cord.exceptions.CordException(
+                message='You must set a title to create a model row.'
+            )
+
+        if features is None:
+            raise cord.exceptions.CordException(
+                message="You must pass a list of feature uid's (hashes) to create a model row."
+            )
+
+        if model is None or model not in [RESNET18, RESNET34, RESNET50, RESNET101, RESNET152, VGG16, VGG19, YOLOV5,
+                                          FASTER_RCNN, MASK_RCNN]:
+            raise cord.exceptions.CordException(
+                message="You must pass a model (resnet18, resnet34, resnet50, resnet101, resnet152, vgg16, vgg19, "
+                        "yolov5, faster_rcnn, mask_rcnn) to create a model row."
+            )
+
+        model_row = ModelRow({
+            'title': title,
+            'description': description,
+            'features': features,
+            'model': model,
+        })
+
+        model = Model({
+            'model_operation': ModelOperations.CREATE.value,
+            'model_parameters': model_row,
+        })
+
+        return self._querier.basic_put(Model, None, payload=model)
+
     def model_inference(self,
                         uid,
                         file_paths=None,
@@ -345,18 +402,20 @@ class CordClientProject(CordClient):
                         iou_thresh=0.3,
                         device="cuda",
                         detection_frame_range=None,
+                        allocation_enabled=False,
                         ):
         """
         Run inference with model trained on the platform.
 
         Args:
             uid: A model_iteration_hash (uid) string.
-            file_paths: List of local file paths to image(s) or video(s) - if running inference on files
-            base64_strings: List of base 64 strings of image(s) or video(s) - if running inference on base64 strings
-            conf_thresh: Confidence threshold (default 0.6)
-            iou_thresh: Intersection over union threshold (default 0.3)
-            device: Device (CPU or CUDA, default is CUDA)
-            detection_frame_range: Detection frame range (for videos)
+            file_paths: List of local file paths to image(s) or video(s) - if running inference on files.
+            base64_strings: List of base 64 strings of image(s) or video(s) - if running inference on base64 strings.
+            conf_thresh: Confidence threshold (default 0.6).
+            iou_thresh: Intersection over union threshold (default 0.3).
+            device: Device (CPU or CUDA, default is CUDA).
+            detection_frame_range: Detection frame range (for videos).
+            allocation_enabled: Object UID allocation (tracking) enabled (disabled by default).
 
         Returns:
             Inference results: A dict of inference results.
@@ -372,7 +431,9 @@ class CordClientProject(CordClient):
         if (file_paths is None and base64_strings is None) or (
                 file_paths is not None and len(file_paths) > 0 and base64_strings is not None and len(
                 base64_strings) > 0):
-            raise Exception("To run model inference, you must pass either a list of files or base64 strings")
+            raise cord.exceptions.CordException(
+                message='To run model inference, you must pass either a list of files or base64 strings.'
+            )
 
         if detection_frame_range is None:
             detection_frame_range = []
@@ -393,15 +454,90 @@ class CordClientProject(CordClient):
                     "base64_str": base64_string.decode('utf-8')  # base64 string to utf-8
                 })
 
-        params = ModelInferenceParams({
+        inference_params = ModelInferenceParams({
             'files': files,
             'conf_thresh': conf_thresh,
             'iou_thresh': iou_thresh,
             'device': device,
             'detection_frame_range': detection_frame_range,
+            'allocation_enabled': allocation_enabled,
         })
 
-        return self._querier.basic_setter(Model, uid, payload=params)
+        model = Model({
+            'model_operation': ModelOperations.INFERENCE.value,
+            'model_parameters': inference_params,
+        })
+
+        return self._querier.basic_setter(Model, uid, payload=model)
+
+    def model_train(self,
+                    uid,
+                    label_rows=None,
+                    epochs=None,
+                    batch_size=24,
+                    weights=None,
+                    device="cuda"
+                    ):
+        """
+        Train a model created on the platform.
+
+        Args:
+            uid: A model_hash (uid) string.
+            label_rows: List of label row uid's (hashes) for training.
+            epochs: Number of passes through training dataset.
+            batch_size: Number of training examples utilized in one iteration.
+            weights: Model weights.
+            device: Device (CPU or CUDA, default is CUDA).
+
+        Returns:
+            A model iteration object.
+
+        Raises:
+            AuthenticationError: If the project API key is invalid.
+            AuthorisationError: If access to the specified resource is restricted.
+            ModelWeightsInconsistentError: If the passed model weights are incompatible with the selected model.
+            ResourceNotFoundError: If no model exists by the specified model_hash (uid).
+            UnknownError: If an error occurs during training.
+        """
+        if label_rows is None:
+            raise cord.exceptions.CordException(
+                message="You must pass a list of label row uid's (hashes) to train a model."
+            )
+
+        if epochs is None:
+            raise cord.exceptions.CordException(
+                message="You must set number of epochs to train a model."
+            )
+
+        if batch_size is None:
+            raise cord.exceptions.CordException(
+                message="You must set a batch size to train a model."
+            )
+
+        if weights is None:
+            raise cord.exceptions.CordException(
+                message="You must select model weights to train a model."
+            )
+
+        if device is None:
+            raise cord.exceptions.CordException(
+                message="You must set a device (cuda or CPU) train a model."
+            )
+
+        training_params = ModelTrainingParams({
+            'label_rows': label_rows,
+            'epochs': epochs,
+            'batch_size': batch_size,
+            'weights': weights,
+            'device': device,
+        })
+
+        model = Model({
+            'model_operation': ModelOperations.TRAIN.value,
+            'model_parameters': training_params,
+        })
+
+        return self._querier.basic_setter(Model, uid, payload=model)
 
     def object_interpolation(self,
                              key_frames,
@@ -445,10 +581,10 @@ class CordClientProject(CordClient):
                     ...,
                 }
 
-            objects_to_interpolate: List of object uid's (hashes) of objects to interpolate
+            objects_to_interpolate: List of object uid's (hashes) of objects to interpolate.
 
         Returns:
-            Interpolation results: Full set of filled frames including interpolated objects
+            Interpolation results: Full set of filled frames including interpolated objects.
 
         Raises:
             AuthenticationError: If the project API key is invalid.
@@ -456,16 +592,18 @@ class CordClientProject(CordClient):
             UnknownError: If an error occurs while running interpolation.
         """
         if len(key_frames) == 0 or len(objects_to_interpolate) == 0:
-            raise Exception("To run object interpolation, you must pass key frames and objects to interpolate")
+            raise cord.exceptions.CordException(
+                message="To run object interpolation, you must pass key frames and objects to interpolate.."
+            )
 
         interpolation_params = ObjectInterpolationParams({
             'key_frames': key_frames,
             'objects_to_interpolate': objects_to_interpolate,
         })
 
-        params = LabelingAlgorithm({
+        algo = LabelingAlgorithm({
             'algorithm_name': INTERPOLATION,
             'algorithm_parameters': interpolation_params,
         })
 
-        return self._querier.basic_setter(LabelingAlgorithm, str(uuid.uuid4()), payload=params)
+        return self._querier.basic_setter(LabelingAlgorithm, str(uuid.uuid4()), payload=algo)
